#
# mbox.rb
#
#   Copyright (c) 1999 Minero Aoki <aamine@dp.u-netsurf.ne.jp>
#
#   This program is free software.
#   You can distribute/modify this program under the terms of
#   the GNU Lesser General Public License version 2 or later.
#

require 'amstd/must'
require 'amstd/bug'
require 'amstd/futils'
require 'amstd/extmod'

require 'tmail/tmail'


module EST

  class CMailset

    Export = %w(
      fetch [] mbox_names
    )

    def initialize
      @set = {}
      @rc = {}
      @inbox = nil
      @draft = nil
      @trash = nil
    end

    def shutdown
      @set.each do |name, mb|
        if Mailbox === mb then
          mb.finalize
        end
      end
    end


    def add_mbox( name, rc )
      rc.must Hash if $DEBUG
      if @rc[name] then
        if @set[name] then
          raise ArgumentError, "realized mbox '#{name}' is re-registered"
        else
          CORE.warn "mbox '#{name}' registered twice"
        end
      end
# puts "mbox added: name=#{name}"
      @rc[name] = rc
    end

    def delete_mbox( name )
      unless @rc[name] then
        raise ArgumentError,
          "not exist mbox '#{name}' was tryed to delete"
      end

      if mb = @set[name] then
        mb.finalize
        @set.delete name
      end
      @rc.delete name
    end


    def []( name )
      name.must String if $DEBUG

      unless ret = @set[ name ] then
        if rc = @rc[ name ] then
          lo = mkloader( rc )
          @set[ name ] = ret = make_real_mbox( name, rc, lo )
        else
          ret = nombox( name )
        end
      end

      ret
    end

    def nombox( name )
      raise ArgumentError, "no such mbox: #{name}"
    end


    class << self
      def sysmb_attr( label )
        label = label.id2name
        module_eval %-
          def #{label}=( name )
            name.must String
            if mb = @rc[ name ] then
              @#{label} = name
            else
              raise ArgumentError,
                "not exist mbox '\#{name}' was given for #{label}"
            end
          end
        -
        module_eval %-
          def #{label}
            self[ @#{label} ]
          end
        -
      end
    end

    sysmb_attr :inbox
    sysmb_attr :draft
    sysmb_attr :trash


    def mbox_names
      ret = @rc.collect {|name,rc| rc['viewname'] || name }
      ret.sort!
      ret
    end


    def mkloader( rc )
      case rc[ 'type' ]
      when 'mh'
        lo = ::TMail::MhLoader.new( rc['path'] )
      else
        bug! 'mbox type is not MH'
      end

      return lo
    end

    def make_real_mbox( name, rc, lo )
      Mailbox.new( name, rc, lo )
    end

  end

  class Mailset < CMailset
  end



  class CMailbox

    include Enumerable

    def initialize( name, rc, lo, *rest )
      name.must String if $DEBUG
      @name   = name.dup.freeze
      @rc     = rc
      @loader = lo

      @array  = []

      do_initialize( *rest )
      load
    end

    def finalize
      do_finalize
    end


    private


    abstract :create_getter  # by interface
    abstract :create_setter  # by interface

    abstract :do_initialize  # by interface
    abstract :do_finalize    # by subclass

    abstract :load           # by subclass


    public


    ###  attributes

    attr :name

    def size
      @array.size
    end
    alias length size


    ###  access/search

    def []( idx )
      return nil unless idx
      @array[idx]
    end
    
    def each( &block )
      @array.each( &block )
    end

    def each_index( &block )
      @array.each_index( &block )
    end

    def each_mail
      each{|port| yield Mail.new( port ) }
    end

    def index( port )
      port.must ::TMail::Port if $DEBUG
      @array.index port
    end

    def rindex( port )
      port.must ::TMail::Port if $DEBUG
      @array.rindex port
    end

    def indexes( *args )
      @array.indexes *args
    end

    def to_a
      @array.dup
    end


    ###  mail handle

    def add( port )
      do_add port
    end

    def moveof( mbox, idx )
      do_moveof mbox, idx
    end

    def remove( idx )
      do_remove idx
    end

    def new_port
      do_new_port
    end
    alias new new_port
    
    def update
      do_update
    end

    def sort( &block )
      @array.sort!( &block )
    end


    private

    abstract :do_add
    # abstract :do_cp
    abstract :do_move_of
    abstract :do_remove
    abstract :do_new_port

  end

  class Mailbox < CMailbox
  end



  class RealMailbox < Mailbox

    def update
      hash = {}
      each{|port| hash[ port ] = true }
      @loader.each do |p|
        p.unread = true if @name == 'inbox'
        add p unless hash.has_key? p
      end
    end


    private

    def load
      CORE.notice 'loading mails ...'

      if entries = load_cache then
        entries.each do |ent|
          port = ent.shift
          add port, ent
        end
      else
puts 'parsing'
        unread = (@name == 'inbox')
        @loader.each do |p|
          p.unread = unread
          add p
        end
      end
    end

    def do_finalize
      save_cache if @loader.can_cache?
    end


    def do_add( port )
      port.must TMail::Port

      port.ref
      @array.push port

      port
    end

    def do_moveof( mbox, idx )
      p = @loader.new_port
# puts "do_mv: from=#{port},to=#{p}"
      port = mbox[idx]
      port.mv p
      p.unread = port.unread
      add p
      mbox.remove idx

      return p
    end

    def do_remove( idx )
      return unless port = @array.delete_at( idx )
      port.unref
      return nil
    end

    def do_new_port
      @loader.new_port
    end
  

    #
    # cache system
    #

    CacheVersion = 2

    def load_cache
      fn = "#{CORE['usrdir']}/cache/#{EST::Interface}/#{@name}"
      entries = nil

      begin
        if File.exist? fn then
p fn
          File.open( fn ) do |f|
            ver = f.gets.downcase.strip
            break unless ver == "est cache file version #{CacheVersion}"
puts 'version ok'

            size = f.gets.strip.to_i
            sign = Marshal.load( f.read size )
            break unless cache_signature == sign
puts 'correct cache'

puts "loading cache: #{@name}"
            entries = Marshal.load( f )
          end
        end
unless entries then
puts "cache miss: #{@name}"
end
      rescue
        CORE.warn 'error in reading cache file (may not be fatal)'
        CORE.warn $!.to_s
        entries = nil
      end

      entries
    end

    def save_cache
puts "saving cache: #{@name}"
      fn = "#{CORE['usrdir']}/cache/#{EST::Interface}/#{@name}"

      begin
        FileUtils.mkdir_p File.dirname(fn)
        File.open( fn, 'w' ) do |f|
          f.puts "est cache file version #{CacheVersion}"

          str = Marshal.dump( cache_signature )
          f.puts str.size.to_s
          f.write str

          ent = collect{|port| port.cache }
          Marshal.dump( ent, f )
        end
      rescue
        $stderr.puts 'error in writing cache file (may not be fatal)'
        $stderr.puts $!.to_s
        if File.file? fn then
          FileUtils.rm_f fn
        end
      end
    end

    # abstract cache_signature()

  end


  class ReferenceError < StandardError ; end


  class VirtualMailbox < Mailbox

    private

    def load
    end

    def do_finalize
    end

    def do_add( port )
      port.must TMail::Port
      @array.push port
      port.ref
    end

    def do_move_of( mbox, idx )
      raise ReferenceError, "cannot move mail to virtual mbox"
    end

    def do_remove( idx )
      port = @array.delete_at( idx )
      port.unref if port
    end

    def do_new_port
      raise ReferenceError, "virtual mbox cannot make new mail"
    end

  end



  class Mail < TMail::Mail

    def port=( p )
      unless @parsed then
        body
      end
      @port.mv p
      @port = p
    end
      
    def sync
      @port.wopen do |wp|
        wp.write to_s
      end
    end

    attr :cache

  end


end   # module EST



# post-defines


module TMail

  class Port

    def initialize
      @refc = 0
      @unread = false
      @cache = nil
    end

    attr :cache,  true
    attr :unread, true

    def ref
      @refc += 1
    end

    def unref
      @refc -= 1
#puts "port#unref, refc=#{@refc}"
      if @refc < 1 then
#puts "delete file"
        rm
      else
#puts "not delete"
      end
    end

  end

  class FilePort

    def _dump( limit )
      @filename
    end

    def FilePort._load( str )
      new( str )
    end

  end

  class Loader

    def can_cache?
      false
    end

  end

  class MhLoader

    def can_cache?
      true
    end

  end

end
